#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  

void LinearComplexity(int M, int n)
{
  int       i, ii, j, d, N;
  int       L, m, N_, state, parity, sign;
  double    p_value, T_, mean;
  char      assignment[35];
  int       K = 6;
  double    pi[7]={0.01047,0.03125,0.12500,0.50000,0.25000,0.06250,0.020833};
  double    nu[7], chi2;
  BitField  *T, *P, *B_, *C;
  
  N = (int)floor(n/M);
  if (((B_ = (BitField*) calloc(M,sizeof(BitField))) == NULL) ||
      ((C  = (BitField*) calloc(M,sizeof(BitField))) == NULL) ||
      ((P  = (BitField*) calloc(M,sizeof(BitField))) == NULL) ||
      ((T  = (BitField*) calloc(M,sizeof(BitField))) == NULL)) {
     printf("Insufficient Memory for Work Space:: Linear Complexity Test\n");
     fflush(stdout);
     if (B_!= NULL) free(B_);
     if (C != NULL) free(C);
     if (P != NULL) free(P);
     if (T != NULL) free(T);
     return;
  }
  else {
     if (LINEAR_COMPLEXITY) {
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],"\tL I N E A R  C O M P L E X I T Y\n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],"\tM (substring length)     = %d\n", M);
        fprintf(stats[16],"\tN (number of substrings) = %d\n", N);
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],
                "        F R E Q U E N C Y                            \n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],
                "  C0   C1   C2   C3   C4   C5   C6    CHI2    P-value\n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        if (n%M != 0) 
           fprintf(stats[16],"\tNote: %d bits were discarded!\n", n%M);
     }
     for(i = 0; i < K+1; i++) nu[i] = 0.00;
     for(ii = 0; ii < N; ii++) {
        for(i = 0; i < M; i++) {
           B_[i].b = 0;
           C[i].b = 0;
           T[i].b = 0;
           P[i].b = 0;
        }
        L = 0;
        m = -1;
        d = 0;
        C[0].b = 1;
        B_[0].b = 1;
        /*
        printf("d = %2d ", d);
        printf("\nL = %2d\nm = %2d\n", L,m);
        printf("N_ = %2d\nS_N => ", N_); 
        for(i = 0; i < M; i++) printf("%d", (int)epsilon[i].b); 
        printf("\nT   => ");
        for(i = 0; i < M; i++) printf("%d", (int)T[i].b); 
           printf("\nC   => ");
        for(i = 0; i < M; i++) printf("%d", (int)C[i].b); 
           printf("\nB   => ");
        for(i = 0; i < M; i++) printf("%d", (int)B[i].b); 
        printf("\n");
        */
        /* DETERMINE LINEAR COMPLEXITY */
        N_ = 0;
        /*for(i = 0; i < M; i++) eps[i].b = epsilon[ii*M+i].b;*/
        while(N_ < M) {
           d = (int)epsilon[ii*M+N_].b;
           /*d = eps[N_].b;*/
           for(i = 1; i <= L; i++)
              /*d += (int)C[i].b*(int)epsilon[N_-i].b;*/
              d += (int)C[i].b*(int)epsilon[ii*M+N_-i].b;
           d = d%2;
           if (d == 1) {
              for(i = 0; i < M; i++) {
                 T[i].b = C[i].b;
                 P[i].b = 0;
              }
              for(j = 0; j < M; j++)
                 if (B_[j].b == 1) P[j+N_-m].b = 1;
              for(i = 0; i < M; i++)
                 C[i].b = (C[i].b + P[i].b)%2;
              if (L <= N_/2) {
                 L = N_ + 1 - L;
                 m = N_;
                 for(i = 0; i < M; i++)
                    B_[i].b = T[i].b;
              }
           }
           N_++;
           /*
           printf("\nLinear Complexity L = %d\n", L);
           printf("--------------------------------------------\n");
           printf("d = %2d ", d);
           printf("\nL = %2d\nm = %2d\n", L,m);
           printf("N_ = %2d\nS_N => ", N_); 
           for(i = 0; i < M; i++) printf("%d", (int)epsilon[i].b); 
           printf("\nT   => ");
           for(i = 0; i < M; i++) printf("%d", (int)T[i].b); 
           printf("\nC   => ");
           for(i = 0; i < M; i++) printf("%d", (int)C[i].b); 
           printf("\nB   => ");
           for(i = 0; i < M; i++) printf("%d", (int)B[i].b); 
           printf("\n");
           printf("%2d\n", L);
           */
        }
        if ((parity = (M+1)%2) == 0) 
           sign = -1;
        else 
           sign = 1;
        mean = M/2. + (9.+sign)/36. - 1./pow(2,M) * (M/3. + 2./9.);
        if ((parity = M%2) == 0) 
           sign = 1;
        else 
           sign = -1;
        T_ = sign * (L - mean) + 2./9.;

        /*printf("L = %5d mean = %f T = %f\n", (int)L,mean,T_);*/

        if (T_ <= -2.5)
           nu[0]++;
        else if (T_ > -2.5 && T_ <= -1.5)
           nu[1]++;
        else if (T_ > -1.5 && T_ <= -0.5)
           nu[2]++;
        else if (T_ > -0.5 && T_ <= 0.5)
           nu[3]++;
        else if (T_ > 0.5 && T_ <= 1.5)
           nu[4]++;
        else if (T_ > 1.5 && T_ <= 2.5)
           nu[5]++;
        else
           nu[6]++;
  
        /*printf("--------------------------------------------\n");
        fprintf(stats[16],"Connection Polynomial:  C(D) = 1 + ");
        for(i = 1; i < M; i++) {
           if ((int)C[i].b != 0) {
              fprintf(stats[16],"D^%d", i);
              if (i < M) fprintf(stats[16]," + ");
           }
        }
        p_value = 0.5*erfc((mean-L)/pow(2.0*variance,0.5));
        if ((ii+1)%100 == 0) printf("%dth substring complete\n", ii+1);
        */
     }
     chi2 = 0.00;
     for(i = 0; i < K+1; i++) 
        fprintf(stats[16],"%4d ",(int)nu[i]);
     for(i = 0; i < K+1; i++)
        chi2 += pow(nu[i]-N*pi[i],2)/(N*pi[i]);
     p_value = igamc(K/2.0,chi2/2.0);
     /*p_value = gammq(K/2.0,chi2/2.0);*/
     if (p_value < ALPHA) {				    /* INTERPRETATION */
        strcpy(assignment,"FAILURE");
        state = 0;
     }
     else {
        strcpy(assignment,"SUCCESS");
        state = 1;
     }
     fprintf(stats[16],"%9.6f%9.6f\n", chi2, p_value);
     fprintf(results[16],"%f\n", p_value);
     fprintf(grid, "%d", state);
     fflush(stats[16]);
     fflush(results[16]);
     fflush(grid);
  }
  free(B_);
  free(P);
  free(C);
  free(T);
  return;
}
