/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                            U T I L I T I E S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"

extern double lgam ( double x );
void byteReverse( ULONG*, int );

void fileBasedBitStreams(char** streamFile)
{ 
  FILE* fp;
  int   mode;

  printf("\n\t\t[0] BITS IN ASCII FORMAT\n");
  printf("\n\t\t[1] HEX DIGITS IN BINARY FORMAT\n");
  printf("\n\t\tSelect input mode:  ");
  scanf("%1d",&mode);
  printf("\n");
  if (mode == 0) {
     if ((fp = fopen(*streamFile, "r")) == NULL) {
        fprintf(output, "BitStream file %s could not be opened.\n",*streamFile);
        printf("ERROR IN FUNCTION fileBasedBitStreams:  file %s ", *streamFile);
		printf("could not be opened.\n");
		fflush(stdout);
		exit(-1);
		}
     readBinaryDigitsInASCIIFormat(fp,*streamFile);
     fclose(fp);
  }
  else if (mode == 1) {
     if ((fp = fopen(*streamFile, "rb")) == NULL) {
        printf("ERROR IN FUNCTION fileBasedBitStreams:  file %s ", *streamFile);
		printf("could not be opened.\n");
		exit(-1);
		}
     readHexDigitsInBinaryFormat(fp);
     fclose(fp);
  }
  return;
}

double Pr(int u, double eta)
{
  int    l;
  double sum, p;

  if (u == 0)
     p = exp(-eta);
  else {
     sum = 0.0;
     for(l = 1; l <= u; l++)
        sum += exp(-eta-u*log(2)+l*log(eta)-lgam(l+1)+lgam(u)-lgam(l)-
                   lgam(u-l+1));
        /*sum += exp(-eta-u*log(2)+l*log(eta)-gammln(l+1)+gammln(u)-gammln(l)-
                   gammln(u-l+1));*/
     p = sum;
  }
  return p;
}

void induceBias(int n)
{
  int i, j, num, n0, n1;
  
  j = 0;
  while (j < floor(n-20)/40) {
     j++;
     num = 40*j-1;
     epsilon[num].b = 0;
     epsilon[num+20].b = 1;
  }
  n0 = 0;
  n1 = 0;
  for(i=0; i < n; i++) {
     if (epsilon[i].b == 0) 
        n0++;
     else 
        n1++;
  }
  fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", n, n0, n1);
  fflush(output);
  return;
}

void nist_test_suite()
{
  int w;

  if ((testVector[0] == 1) || (testVector[1] == 1)) 
     Frequency(tp.n);

  if ((testVector[0] == 1) || (testVector[2] == 1)) 
     BlockFrequency(tp.blockFrequencyBlockLength, tp.n);

  if ((testVector[0] == 1) || (testVector[3] == 1)) 
     CumulativeSums(tp.n);

  if ((testVector[0] == 1) || (testVector[4] == 1)) 
     Runs(tp.n); 

  if ((testVector[0] == 1) || (testVector[5] == 1)) 
     LongestRunOfOnes(tp.n);

  if ((testVector[0] == 1) || (testVector[6] == 1)) 
     Rank(tp.n);

  if ((testVector[0] == 1) || (testVector[7] == 1)) 
     DiscreteFourierTransform(tp.n);

  if ((testVector[0] == 1) || (testVector[8] == 1)) 
     NonOverlappingTemplateMatchings(tp.nonOverlappingTemplateBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[9] == 1))
     OverlappingTemplateMatchings(tp.overlappingTemplateBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[10] == 1))
     Universal(tp.universalBlockLength, tp.universalNumberInitializationSteps,
               tp.n);

  if ((testVector[0] == 1) || (testVector[11] == 1)) 
     ApproximateEntropy(tp.approximateEntropyBlockLength, tp.n);

  if ((testVector[0] == 1) || (testVector[12] == 1)) 
     RandomExcursions(tp.n);

  if ((testVector[0] == 1) || (testVector[13] == 1)) 
     RandomExcursionsVariant(tp.n);

  if ((testVector[0] == 1) || (testVector[14] == 1))
     Serial(tp.serialBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[15] == 1))
     w = LempelZivCompression(tp.n);

  if ((testVector[0] == 1) || (testVector[16] == 1))
     LinearComplexity(tp.linearComplexitySequenceLength,tp.n);

  fprintf(grid, "\n");
  fflush(grid);
  return;
}

void readBinaryDigitsInASCIIFormat(FILE* fp, char* streamFile)
{
  int i, j, n0, n1, bitsRead;
  unsigned int bit;	

  if ((epsilon = (BitField*) calloc(tp.n,sizeof(BitField))) == NULL) {
     printf("BITSTREAM DEFINITION:  Insufficient memory available.\n");
     printf("Statistical Testing Aborted!\n");
     return;
  }
  else {
     printf("\n\t\tStatistical Testing In Progress.........\n");   
     for(i = 0; i < tp.numOfBitStreams; i++) {
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
        for(j = 0; j < tp.n; j++) {
           if (fscanf(fp, "%1d", &bit) == EOF) {
      	       printf("ERROR:  Insufficient data in file %s.  ", streamFile);
	       printf("%d bits were read.\n", bitsRead);
  	       fflush(stdout);
    	       fclose(fp);
	       return;
           }
           else {
              bitsRead++;
              if (bit == 0) 
                 n0++;
              else 
                 n1++;
              epsilon[j].b = bit;
           }
        }
        fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
        nist_test_suite();
     }
     free(epsilon);
  }
  return;
}

void readHexDigitsInBinaryFormat(FILE* fp)
{
  unsigned bit;				/* BINARY DIGIT        */
  unsigned long* buffer;
  unsigned long  c, displayMask = 1 << 31;
  int            i, k, n0, n1, bitsRead, numRead;
 
  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL)) {
     printf("BITSTREAM DEFINITION:  Insufficient memory available.\n");
     fflush(stdout);
     return;
  }
  if (((buffer = (unsigned long*) calloc((unsigned int)floor(tp.n/32)+1,
	sizeof(unsigned long))) == NULL)) {
     printf("BITSTREAM DEFINITION:  Insufficient memory available.\n");
     return;
  }
  else {
     printf("\n\t\tStatistical Testing In Progress.........\n");   
     for(k = 0; k < tp.numOfBitStreams; k++) {
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
        numRead = (int)floor(tp.n/32);
        if ((tp.n)%32 != 0) numRead++;
        for(i = 0; i < numRead; i++) {
      	   fread(buffer,sizeof(unsigned long),1,fp);
#ifdef LITTLE_ENDIAN
	    byteReverse( buffer, 4 );
#endif
	    for(c = 1; c <= 32; c++) {
	      if (buffer[0] & displayMask) {
             bit = 1;
			 n1++;
			 }
	      else {
			 bit = 0;
			 n0++;
			 }
	      buffer[0] <<= 1;
	      epsilon[bitsRead].b = bit;
	      bitsRead++;
	      if (bitsRead == tp.n) break;
	      }
       }
       fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
       fflush(output);

       nist_test_suite();

     }
     free(buffer);
     free(epsilon);
  }
  return;
}
