#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/config.h"
#include "../include/utilities1.h"
#include "../include/utilities2.h"
#include "../include/generators1.h"
#include "../include/generators2.h"

/* function declarations from generator3.h */
void SHA1();
void ANSI();
void DES();

int generatorOptions(char** streamFile)
{
  char file[200];
  int  option = NUMOFGENERATORS+1; 

  while ((option < 0) || (option > NUMOFGENERATORS)) {
     option = displayGeneratorOptions();
     switch(option) {
        case 0 :  printf("\t\tUser Prescribed Input File: ");
                  scanf("%s", file);
				  *streamFile = (char*)calloc(200,sizeof(char));
				  sprintf(*streamFile,"%s",file);
                  break;
        #if (INCLUDE_GENERATORS == 1)
        case 1 :  *streamFile = "G using SHA-1";
		  break;
        case 2 :  *streamFile = "Linear-Congruential";
		  break;
        case 3 :  *streamFile = "Blum-Blum-Shub";
		  break;
        case 4 :  *streamFile = "Micali-Schnorr";
		  break;
        case 5 :  *streamFile = "Modular-Exponentiation";
		  break;
        case 6 :  *streamFile = "Quadratic-Congruential-1";
		  break;
        case 7 :  *streamFile = "Quadratic-Congruential-2";
		  break;
        case 8 :  *streamFile = "Cubic-Congruential";
		  break;
        case 9 :  *streamFile = "XOR";  
		  break;
        case 10:  *streamFile = "ANSI-X9.17";
		  break;
        case 11 : *streamFile = "G using DES";
		  break;

        /* INTRODUCE NEW PRNG NAMES HERE */
        /*
        case 12:  *streamFile = "myNewPRNG";
		  break;
        */

        #endif
        default:  printf("Error:  Try again!\n");
                  break;
     }
  }
  return option;
}

void invokeTestSuite(int option, char* streamFile)
{
  fprintf(output,"_________________________________________________");
  fprintf(output,"_______________________________\n\n");
  fprintf(output,"\t\tFILE = %s\t\tALPHA = %6.4f", streamFile, ALPHA);
  fprintf(output,"\n________________________________________________");
  fprintf(output,"________________________________\n\n");
  fflush(output);
  if (option != 0) printf("\n\t\tStatistical Testing In Progress.........\n");
  switch(option) {
     case 0 :  fileBasedBitStreams(&streamFile);
	       break;
     #if (INCLUDE_GENERATORS == 1)
     case 1 :  SHA1(); 
               break;
     case 2 :  lcg();
	       break;
     case 3 :  bbs();
	       break;
     case 4 :  micali_schnorr();
	       break;
     case 5 :  modExp();
	       break;
     case 6 :  quadRes1();
	       break;
     case 7 :  quadRes2();
	       break;
     case 8 :  cubicRes();
	       break;
     case 9 :  exclusiveOR();
	       break;
     case 10:  ANSI();
               break;
     case 11:  DES(); 
	       break;
     #endif

     /* INTRODUCE NEW PSEUDO RANDOM NUMBER GENERATORS HERE */

     default:  printf("Error in invokeTestSuite!\n");
	       break;
  }
  printf("\n\t\tStatistical Testing Complete!!!!!!!!!!!!\n");
  return;
}

void openOutputStreams(int option)
{
  int   i, numOfBitStreams;
  /*  char  directory[150]; */
  char  statsDir[200];
  char  resultsDir[200];
  char  generatorDir[NUMOFGENERATORS][20] = {"AlgorithmTesting","G-SHA-1", 
        "LCG", "BBS", "MS", "MODEXP", "QCG1", "QCG2","CCG", "XOR", "ANSI-X9.17", 
        "G-DES"};
  char  testNames[NUMOFTESTS+1][50] = {" ", "frequency", "block-frequency", 
	"cumulative-sums", "runs", "longest-run", "rank", "fft", 
        "nonperiodic-templates", "overlapping-templates", "universal", "apen", 
        "random-excursions", "random-excursions-variant", "serial",
        "lempel-ziv", "linear-complexity"};
  int   numOfOpenFiles = 0;
  
  for(i = 1; i <= NUMOFTESTS; i++) {
     if (testVector[i] == 1) {
        sprintf(statsDir, "experiments/%s/%s/stats", generatorDir[option],
		testNames[i]);
        sprintf(resultsDir, "experiments/%s/%s/results", generatorDir[option],
		testNames[i]);
        if (DISPLAY_OUTPUT_CHANNELS) {
           printf("*STATS* DIR => %s\n", statsDir);
		   printf("RESULTS DIR => %s\n", resultsDir);
           fflush(stdout);
        }
        if ( (stats[i] = fopen(statsDir, "w")) == NULL ) {	/* STATISTICS LOG */
           printf("ERROR: LOG FILES COULD NOT BE OPENED.\n");
           printf("       MAX # OF OPENED FILES HAS BEEN REACHED = %d\n", 
                  numOfOpenFiles);
           printf("-OR-   THE OUTPUT DIRECTORY DOES NOT EXIST.\n");
           fflush(stdout);
           exit(-1);
		   }
		else
           numOfOpenFiles++;
        if ( (results[i] = fopen(resultsDir, "w")) == NULL ) {	/* P_VALUES LOG   */
           printf("ERROR: LOG FILES COULD NOT BE OPENED.\n");
           printf("       MAX # OF OPENED FILES HAS BEEN REACHED = %d\n", 
                  numOfOpenFiles);
           printf("-OR-   THE OUTPUT DIRECTORY DOES NOT EXIST.\n");
           fflush(stdout);
           exit(-1);
		   }
		else
           numOfOpenFiles++;
     }
  }
  printf("\n\t\tHow many bitstreams should be generated? ");
  scanf("%d", &numOfBitStreams);
  tp.numOfBitStreams = numOfBitStreams;
  return;
}

int displayGeneratorOptions()
{
  int option = 0;
   
  #if (INCLUDE_GENERATORS == 1)
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\t\t     G E N E R A T O R  O P T I O N S \n");
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\t[00] Input File\t\t\t[01] G Using SHA-1\n");
  printf("\t[02] Linear Congruential\t[03] Blum-Blum-Shub\n");
  printf("\t[04] Micali-Schnorr\t\t[05] Modular Exponentiation\n");
  printf("\t[06] Quadratic Congruential I\t[07] Quadratic Congruential II\n");
  printf("\t[08] Cubic Congruential\t\t[09] XOR\n");
  printf("\t[10] ANSI X9.17 (3-DES)\t\t[11] G Using DES\n");
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\tOPTION ----> "); scanf("%d",&option);
  #endif
  return option;
}

void chooseTests()
{
  int i;
   
  displayTests();
  printf("\n\n\n");
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\t\t\t     INSTRUCTIONS\n");
  printf("\tEnter 0 if you DO NOT want to apply all of the\n");
  printf("\tstatistical tests to each sequence and 1 if you DO. \n");
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\tEnter Choice:  ");
  scanf("%d",&testVector[0]);
  printf("\n");
  if (testVector[0] == 1)
     for(i=1; i <= NUMOFTESTS; i++) testVector[i] = 1;
  else {
     displayTests();
     printf("________________________________________");
     printf("________________________________________\n\n");
     printf("\t\t\t     INSTRUCTIONS\n");
     printf("\tEnter a 0 or 1 to indicate whether or not the numbered\n");
     printf("\tstatistical test should be applied to each sequence.  \n");
     printf("\tFor example, 1111111111111111 applies every test to \n");
     printf("\teach sequence.\n");
     printf("________________________________________");
     printf("________________________________________\n");
     printf("\t\t\t    1234567891111111\n");
     printf("\t\t\t             0123456\n");
     printf("\t\t\t    ");
     for (i = 1; i <= NUMOFTESTS; i++) 
         scanf("%1d", &testVector[i]);
     printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  }
  return;
}

void displayTests()
{
   printf("________________________________________");
   printf("________________________________________\n\n");
   printf("\t\t      S T A T I S T I C A L  T E S T S\n");
   printf("________________________________________");
   printf("________________________________________\n\n");
   printf("   [01] Frequency\t\t\t");
   printf("   [02] Block Frequency\n");
   printf("   [03] Cumulative Sums\t\t\t");
   printf("   [04] Runs\n");
   printf("   [05] Longest Run of Ones\t\t");
   printf("   [06] Rank\n");
   printf("   [07] Discrete Fourier Transform\t");
   printf("   [08] Nonperiodic Template Matchings\n");
   printf("   [09] Overlapping Template Matchings\t");
   printf("   [10] Universal Statistical\n");
   printf("   [11] Approximate Entropy\t\t");
   printf("   [12] Random Excursions\n");
   printf("   [13] Random Excursions Variant\t");
   printf("   [14] Serial\n");
   printf("   [15] Lempel-Ziv Complexity\t\t");
   printf("   [16] Linear Complexity\n");
   /*printf("   [17] Overlapping Pairs Sparse Occupancy\n");*/
   return;
}

void fixParameters()
{
  if (testVector[2] == 1) {
     printf("\tEnter Block Frequency Test block length:  "); 
     scanf("%d",&tp.blockFrequencyBlockLength);
  }
  if (testVector[8] == 1) {
     printf("\tEnter NonOverlapping Template Test Block Length:  "); 
     scanf("%d", &tp.nonOverlappingTemplateBlockLength);
  }
  if (testVector[9] == 1) {
     printf("\tEnter Overlapping Template Test Block Length:  "); 
     scanf("%d", &tp.overlappingTemplateBlockLength);
  }
  if (testVector[10] == 1) {
     printf("\tEnter Universal Test Block Length:  "); 
     scanf("%d", &tp.universalBlockLength);
     printf("\tEnter Universal Test Number Of Initialization Steps:  "); 
     scanf("%d", &tp.universalNumberInitializationSteps);
  }
  if (testVector[11] == 1) {
     printf("\tEnter Approximate Entropy Test Block Length:  "); 
     scanf("%d", &tp.approximateEntropyBlockLength);
  }
  if (testVector[14] == 1) {
     printf("\tEnter Serial Test Block Length:  "); 
     scanf("%d", &tp.serialBlockLength);
  }
  if (testVector[16] == 1) {
     printf("\tEnter Linear Complexity Test Subsequence Length:  "); 
     scanf("%d", &tp.linearComplexitySequenceLength);
  }
  return;
}

int convertToBits(BYTE* x, int n, int* n0, int* n1, int* bitsRead, int offset, 
                  int num32BitIntegers)
{
  int      i, bit, done;
  BYTE     *T;
  unsigned long  c, displayMask = 1 << 31;
  unsigned long A;

  T = (BYTE*) calloc(9,sizeof(BYTE));
  done = 0;
  for(i = 0; i < num32BitIntegers; i++) {
     sprintf(T,"%02x%02x%02x%02x",x[4*i+offset],x[4*i+offset+1],
             x[4*i+offset+2], x[4*i+offset+3]);
     A = strtoul(T,(char**)NULL,16);
     for(c = 1; c <= 32; c++) {
        if (A & displayMask)
           bit = 1;
        else
           bit = 0;
        A <<= 1;
        if (bit == 0) 
           (*n0)++;
        else 
           (*n1)++;
        epsilon[(int)*bitsRead].b = bit;
        (*bitsRead)++;
        if ((int)*bitsRead == n) {
	   done = 1;
           break;
        }
    }
    if (done) break;
 }
 free(T);
 return done;
}
